from os.path import join
from sysconfig import get_config_var

from cffi import FFI

ffibuilder = FFI()
ffibuilder.set_unicode(False)

header = r"""
typedef struct LikNormMachine LikNormMachine;
enum Lik {
    BERNOULLI,
    BINOMIAL,
    POISSON,
    EXPONENTIAL,
    GAMMA,
    GEOMETRIC
};

LikNormMachine* create_machine(int);
void apply1d(LikNormMachine *, enum Lik, size_t, double *, double *,
             double *, double *, double *, double *);
void apply2d(LikNormMachine *, enum Lik, size_t, double *, double *,
             double *, double *, double *, double *, double *);
void destroy_machine(LikNormMachine *);
"""

ffibuilder.cdef(header)

ffibuilder.set_source(
    "liknorm.machine_ffi",
    header + r"""
#include "liknorm/liknorm.h"

LikNormMachine *create_machine(int n) { return liknorm_create_machine(n); }
void destroy_machine(LikNormMachine *machine) {
  liknorm_destroy_machine(machine);
}

typedef void lik1d(LikNormMachine *, double);
typedef void lik2d(LikNormMachine *, double, double);

void *set_lik[] = {liknorm_set_bernoulli, liknorm_set_binomial,
                   liknorm_set_poisson,   liknorm_set_exponential,
                   liknorm_set_gamma,     liknorm_set_geometric};

void apply1d(LikNormMachine *machine, enum Lik lik, size_t size, double *x,
             double *tau, double *eta, double *log_zeroth, double *mean,
             double *variance) {
  size_t i;
  for (i = 0; i < size; ++i) {
    ((lik1d *)set_lik[lik])(machine, x[i]);
    liknorm_set_prior(machine, tau[i], eta[i]);
    liknorm_integrate(machine, log_zeroth + i, mean + i, variance + i);
  }
}

void apply2d(LikNormMachine *machine, enum Lik lik, size_t size, double *x0,
             double *x1, double *tau, double *eta, double *log_zeroth,
             double *mean, double *variance) {
  size_t i;
  for (i = 0; i < size; ++i) {
    ((lik2d *)set_lik[lik])(machine, x0[i], x1[i]);
    liknorm_set_prior(machine, tau[i], eta[i]);
    liknorm_integrate(machine, log_zeroth + i, mean + i, variance + i);
  }
}""",
    libraries=['liknorm'],
    library_dirs=[join(get_config_var('prefix'), 'lib')],
    include_dirs=[join(get_config_var('prefix'), 'include')])

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



from numpy import asarray, empty, float64
from numpy.random import RandomState
from numpy.testing import assert_allclose, assert_equal

from liknorm import LikNormMachine


def test_sizeof_double():
    from liknorm.machine_ffi import ffi
    assert_equal(ffi.sizeof("double"), 8)


def test_alignof_double():
    from liknorm.machine_ffi import ffi
    assert_equal(ffi.alignof("double"), 8)


def test_liknormmachine():
    machine = LikNormMachine('binomial', 500)
    random = RandomState(0)

    ntrials = random.randint(1, 100, 5)
    nsuccesses = [random.randint(0, i + 1) for i in ntrials]

    ntrials = asarray(ntrials)
    nsuccesses = asarray(nsuccesses)

    y = (nsuccesses, ntrials)
    ceta = random.rand(5)
    ctau = random.randn(5) * ceta

    lmom0 = empty(5, dtype=float64)
    hmu = empty(5, dtype=float64)
    hvar = empty(5, dtype=float64)

    machine.moments(y, ceta, ctau,
                    {'log_zeroth': lmom0,
                     'mean': hmu,
                     'variance': hvar})

    assert_allclose(
        lmom0, [
            -3.478250e+00, -6.179203e+00, -4.473831e+00, -6.086356e+08,
            -7.042068e+00
        ],
        atol=1e-7)
    assert_allclose(
        hmu, [
            1.952541e+00, -1.351837e+00, -6.763551e-01, 1.901923e+07,
            -1.438112e+00
        ],
        rtol=1e-5,
        atol=1e-5)
    assert_allclose(
        hvar,
        [2.008703e-01, 1.257172e-01, 6.619332e-02, 9.629856e+11, 9.355299e-02],
        atol=1e-7)

    machine.finish()



from numpy import all as npall
from numpy import asarray, float64, isfinite

from .machine_ffi import ffi, lib
from .machine_ffi.lib import apply1d, apply2d, create_machine, destroy_machine


def ptr(a):
    return ffi.cast("double *", a.ctypes.data)


class LikNormMachine(object):
    r"""Moments of ExpFam times Normal distribution.

    Example
    -------

    .. doctest::

        >>> from numpy import empty, float64
        >>> from numpy.random import RandomState
        >>> from liknorm import LikNormMachine
        >>>
        >>> machine = LikNormMachine('bernoulli')
        >>> random = RandomState(0)
        >>> outcome = random.randint(0, 2, 5)
        >>> tau = random.rand(5)
        >>> eta = random.randn(5) * tau
        >>>
        >>> log_zeroth = empty(5, dtype=float64)
        >>> mean = empty(5, dtype=float64)
        >>> variance = empty(5, dtype=float64)
        >>>
        >>> moments = {'log_zeroth': log_zeroth, 'mean': mean,
        ...            'variance': variance}
        >>> machine.moments(outcome, eta, tau, moments)
        >>>
        >>> print('%.3f %.3f %.3f' % (log_zeroth[0], mean[0], variance[0]))
        -0.671 -0.515 0.946
    """

    def __init__(self, likname, npoints=500):
        self._likname = likname
        self._machine = create_machine(npoints)
        self._lik = getattr(lib, likname.upper())
        if likname.lower() == 'binomial':
            self._apply = apply2d
        else:
            self._apply = apply1d

    def finish(self):
        destroy_machine(self._machine)

    def moments(self, y, eta, tau, moments):
        r"""First three moments of ExpFam times Normal distribution.

        Args:
            likname (string): likelihood name.
            y (array_like): outcome.
            eta (array_like): inverse of the variance (1/variance).
            tau (array_like): mean times eta.
            moments (dict): log_zeroth, mean, and variance result.
        """

        size = len(moments['log_zeroth'])
        if not isinstance(y, (list, tuple)):
            y = (y, )

        y = tuple(asarray(yi, float64) for yi in y)
        tau = asarray(tau, float64)
        eta = asarray(eta, float64)

        args = y + (tau, eta, moments['log_zeroth'], moments['mean'],
                    moments['variance'])

        self._apply(self._machine, self._lik, size, *(ptr(a) for a in args))

        if not npall(isfinite(moments['log_zeroth'])):
            raise ValueError("Non-finite value found in _log_zeroth_.")

        if not npall(isfinite(moments['mean'])):
            raise ValueError("Non-finite value found in _mean_.")

        if not npall(isfinite(moments['variance'])):
            raise ValueError("Non-finite value found in _variance_.")



from os import chdir, getcwd
from os.path import abspath, basename, dirname, realpath


def test(verbose=True):
    r"""Run tests to verify this package's integrity.

    Parameters
    ----------
    verbose : bool
        ``True`` to show diagnostic. Defaults to ``True``.

    Returns
    -------
    int
        Exit code: ``0`` for success.
    """

    pkgname = basename(dirname(realpath(__file__)))

    p = __import__(pkgname).__path__[0]
    src_path = abspath(p)
    old_path = getcwd()
    chdir(src_path)

    try:
        return_code = __import__('pytest').main(['-q', '--doctest-modules'])
    finally:
        chdir(old_path)

    if return_code == 0:
        print("Congratulations. All tests have passed!")

    return return_code



from __future__ import absolute_import as _

import _cffi_backend as _

from .machine import LikNormMachine
from .testit import test

__name__ = "liknorm"
__version__ = "1.0.12"
__author__ = "Danilo Horta"
__author_email__ = "horta@ebi.ac.uk"

__all__ = [
    "__name__", "__version__", "__author__", "__author_email__", "test",
    "LikNormMachine"
]
